class Job:
    """
    Create job with process ID, size, and status.
    """
    def __init__(self, pid, memsize, status='inactive'):
        self._pid = pid
        self._memsize = memsize
        self._state = status
    
    def __str__(self):
        return f'jobid: {self._pid} | size: {self._memsize:>5} | status: {self._state}'
    
class Memory:
    """
    Create memory address space of desired size and segment(s).
    """
    def __init__(self, size, blocks=[]):
        self._size = size
        self._blocks = blocks
    
    def create_segments(self, segments):
        self._blocks.extend(segments)

    def allocated(self):
        self._allmem = sum([b._size for b in self._blocks])
        return self._allmem

    def unallocated(self):
        self._unallocated = self._size - self._allmem
        return self._unallocated
    
    def busy(self):
        self._busy = sum([b._size for b in self._blocks if not b._state == 'free'])
        return self._busy

    def free(self):
        self._free = self._allmem - self._busy
        return self._free

    def __repr__(self):
        return f'''\nLogical Memory Space
    {"allocated":<12}: {self.allocated()}
    {"unallocated":<12}: {self.unallocated()}
    {"busy":<12}: {self.busy()}
    {"free":<12}: {self.free()}'''


class Scheduler:
    """
    Initialise the memory scheduler, and process jobs with desired policy.
    """
    def __init__(self, memspace):
        self._memory = memspace
    
    def run_firstfit(self, queue):
        print('\n[*] Initialising first fit memory allocation algorithm...')
        print(f'[!] memory partitions: {len(self._memory._blocks)}')
        print(f'[!] jobs in queue: {len(queue)}')
        for j in queue:
            for b in self._memory._blocks:
                if not j._memsize > b._size and b._state == 'free':
                    j._state = 'active'
                    b._state = 'busy'
                    b._jobid = j._pid
                    b._memaddr = hex(sum([p._size for p in self._memory._blocks if p._bid < b._bid]))
                    print(f'[+] {str(j):<40} >>  {b._memaddr:<6}  [block {b._bid}]')
                    break
            else:
                j._state = 'waiting'
                print(f'[-] {str(j):<40}')
    
    def run_bestfit(self, queue):
        print('\n[*] Initialising best fit memory allocation algorithm...')
        print(f'[!] memory partitions: {len(self._memory._blocks)}')
        print(f'[!] jobs in queue: {len(queue)}')
        for j in queue:
            for b in sorted(self._memory._blocks, key=lambda x: x._size):
                if b._size >= j._memsize and b._state == 'free':
                    j._state, b._state, b._jobid = 'active', 'busy', j._pid
                    b._memaddr = hex(sum([p._size for p in self._memory._blocks if p._bid < b._bid]))
                    print(f'[+] {str(j):<40} >>  {b._memaddr:<6}  [block {b._bid}]')
                    break
            else:
                j._state = 'waiting'
                print(f'[-] {str(j):<40}')
        # for j in queue:
        #     try:
        #         bestfit = min([b for b in self._blocks if j._memsize <= b._size and b._state == 'free'], key=lambda x: x._size)
        #         j._state, bestfit._state, bestfit._jobid = 'active', 'busy', j._pid
        #         bestfit._memaddr = hex(sum([b._size for b in self._blocks if b._bid < bestfit._bid], 1))
        #         print(f'[+] {str(j):<40} >>  {bestfit._memaddr:<6}  [block {bestfit._bid}]')
        #     except ValueError:
        #         j._state = 'waiting'
        #         print(f'[-] {str(j):<40}')
        # for job in queue:
        #     bestfit = None
        #     for block in self._blocks:
        #         if not job._memsize > block._size and block._state == 'free' and (not bestfit or block._size < bestfit._size):
        #             bestfit = block
        #     if bestfit:
        #         job._state = 'active'
        #         bestfit._state = 'busy'
        #         bestfit._jobid = job._pid        
        #     else:
        #         job._state = 'waiting'


class MemoryBlock:
    """
    Create memory segments of desired size for where to load processes.
    """
    def __init__(self, blockid, size, status='free', jobid=None):
        self._bid = blockid
        self._size = size
        self._state = status
        self._jobid = jobid
        self._memaddr = None

    def __repr__(self):
        return f'''\nMemory Block
    {"blockid":<10}: {self._bid}
    {"blocksize":<10}: {self._size}
    {"status":<10}: {self._state}
    {"jobid":<10}: {self._jobid}
    {"address":<10}: {self._memaddr}'''
